pub mod types;

use tool_helper::{Error, Input};
use std::io::BufRead;
use types::{Content, Section};

pub fn scan(input: Input) -> Result<Vec<Section>, Error> {
    process(input.lines().map(|l| l.unwrap()).into_iter())?;

    Err(Error::not_implemented("readmap::scan"))
}

fn process<F: std::iter::Iterator<Item=String>>(mut line_iter:F) -> Result<(), Error> {
    while let Some(mut line) = line_iter.next() {
        loop {
            if line.starts_with(".") {
                line = process_section(line, &mut line_iter)?;
            }

            else {
                break;
            }
        }
    }

    Ok(())
}

fn process_section<F: std::iter::Iterator<Item=String>>(line: String, line_iter:&mut F) -> Result<String, Error> {
    let mut next_line_closure = || {
        line_iter.next().ok_or(Error::from_str("Unexpected end of file"))
    };
    let (mut section, line) = read_section(line, &mut next_line_closure)?;

    println!("Section: {} @{:?} {:?} \"{:?}\"", section.name, section.adr, section.size, section.file);
    process_subsection(&mut section, line, &mut next_line_closure)
}

fn process_subsection<F:FnMut()-> Result<String, Error>>(section: &mut Section, line: Option<String>, next_line: &mut F) -> Result<String, Error> {
    fn push_sub_section(section: &mut Section, sub_section: Option<Section>) {
        if let Some(sub_section) = sub_section {
            section.content.push(Content::Section(sub_section));
        }
    }
    let mut line = {
        if let Some(line) = line {
            line
        }

        else {
            next_line()?
        }
    };

    let mut next_line_closure = || -> Result<String, Error> {
        if let Ok(line) = next_line() {
            return Ok(line);
        }

        else {
            // EOF
            return Ok(String::from(""));
        }
    };

    let mut sub_section = None;

    loop {
        if line.starts_with(" .") {
            push_sub_section(section, sub_section);
            let (section, new_line) = read_section(line, &mut next_line_closure)?;

            println!("\t>>>>> {:?} {:?} {:?} {:?}", section.name, section.adr, section.size, section.file);
            sub_section = Some(section);

            if let Some(new_line) = new_line {
                line = new_line;
                continue;
            }
        }

        else if line.is_empty() {
            push_sub_section(section, sub_section);
            return Ok(line);
        }

        line = next_line_closure()?;
    }
}

fn read_section<F:FnMut()-> Result<String, Error>>(mut line: String, next_line: &mut F) -> Result<(Section, Option<String>), Error> {
    let mut split_line = line.split_whitespace();
    let name           = split_line.next().ok_or(Error::from_str("Couldn't locate section name"))?.to_string();
    let need_new_line  = name.chars().count() > 16;

    if need_new_line {
        line       = next_line()?;
        split_line = line.split_whitespace();
    }

    let adr = read_as::<u64>(split_line.next());
    if adr.is_none() {
        return Ok((Section::new(name, None, None, None), {
            if need_new_line {
                Some(line)
            }

            else {
                None
            }
        }));
    }

    let size = read_as::<usize>(split_line.next());
    let file = {
        if let Some(file) = split_line.next() {
            Some(file.to_string())
        }

        else {
            None
        }
    };
    Ok((Section::new(name, adr, size, file), None))
}

fn read_as<F: num_traits::Num>(str: Option<&str>) -> Option<F> {
    if let Some(str) = str {
        F::from_str_radix(str.trim_start_matches("0x"), 16).ok()
    }

    else {
        None
    }
}